module HatTrace (
 HatTrace,      -- type, representing a redex trail file
 HatNode,       -- type, representing a node within the redex trail
		-- structure of HatTrace and HatNode not revealed

 HatNodeType(   -- type: information about what a node within the Redex
		-- Trail contains
	      HatApplNode,HatConstantNode,
              HatSAT_ANode,HatSAT_BNode,HatSAT_CNode,
	      HatHiddenNode,
	      HatProjNode,HatConstrNode,HatIdentNode,HatCaseNode,
	      HatLambdaNode,HatDummyNode,
	      HatIntNode,HatCharNode,HatIntegerNode,HatRationalNode,
              HatFloatNode, HatDoubleNode,HatCStringNode,
	      HatIfNode,HatGuardNode,HatContainerNode,HatModuleNode,
              HatInvalidNode),

 HatApplType(   -- type: type of application node
	      HatApplied,HatBlackholed,HatComplete,HatValue),

 HatInfixType(  -- type: infixtype of an application node. 
	      HatInfix,HatInfixR,HatInfixL, -- constructors with priority
	      HatNoInfix),                  -- constructor with no priority

 HatSourceRef(..),             -- type for values returned as source reference
	  -- (HatNoSourceRef,  -- no reference given
	  --  HatSourceRef),   -- reference given: row, column, moduleName,
			       -- moduleFile, moduletrusted are exported

 HatRep( -- class HatRep, for all structure representing Hat redexes
	toHatNode,    -- convert to type HatNode
	toUndefined,  -- return an undefined node of the same class instance
	hatNodeType,  -- return type of redex trail node
	hatParent,    -- return parent of redex trail node
	hatApplFun,   -- return node for function of an application/constant node
	hatApplArgs,  -- get list of arguments of an application node
	hatInfix,     -- get infix priority of an constructor/identifier node
	hatResult,    -- get result of an application (type complete only)
	hatName,      -- get name of a name/identifier/constructor node
	hatProjValue, -- return projected value of projection/SATA/SATB/SATC node
        hatSourceRef,    -- get information about source position of an application/
                         -- identifier/constructor node
	hatValueInt,     -- get Int value of an HatIntNode
	hatValueChar,    -- get Char value of an HatCharNode
	hatValueInteger, -- get Integer value of an HatIntegerNode
	hatValueRational,-- get Rational value of an HatRationalNode
	hatValueFloat,   -- get Float value of an HatFloatNode
	hatValueDouble,  -- get Double value of an HatDoubleNode
	hatValueString   -- get String value of an HatCStringNode
       ),

 openTrace,	-- open a trace file, return a value of type HatTrace 
 hatTraceName,	-- return file name of file represented by trace
 hatMain,	-- get node representing the "main" caf only
 hatTraceFirst,	-- get first node within HatTrace file
 hatTraceNext,	-- get node immediately after a node

 isValidNode,	-- test if node is valid
 isInvalidNode,	-- test if node is invalid
 hatInvalidNode,-- return an invalid hat node

 toRemoteRep,	-- convert any HatRep to command-line represenation
		-- (in order to pass to another hat tool)
 fromRemoteRep,	-- convert command-line representation to node (in order to
		-- accept an argument from another hat tool)

 -- functions for convenience
 hatIsCAF,     -- check, whether a node represents a CAF or not
 hatIsTopLevel,-- check, whether an application/name/identifier is top-level
 hatIsTrusted, -- check, whether an application/name/identifier is trusted
 hatLeftmost,  -- get leftmost outermost constructor/identifier of application

 -- other views on the redex trail

 detect,       -- return list of EDT children of a node


 observe,      -- make observation: observe a function
 observeSrc,   -- make observation: observe a source reference
 observables,  -- make observation: observe all observable identifiers/modules

 ObserveResult(-- type: type for the result for "observe"
	       IdentNotFound,TopIdentNotFound,Found),
 isIdentNotFound,    -- returns true, if ObserveResult is IdentNotFound
 isTopIdentNotFound, -- returns true, if ObserveResult is TopIdentNotFound
 isFound,            -- returns true, if ObserveResult is Found
 fromFound,          -- get list of observations from "Found" result
 
 -- The following pretty printing functions use C for formatting.
 -- Pretty printing by Haskell is available in the "HatExpression" module!

 hatCExpressionStr,   -- get pretty print for expression, using C's pretty printer
 printCExpression,    -- print expression, using C's pretty printer
 printCReduction,     -- print reduction (lhs = rhs) using C's pretty printer
 printCReductionList  -- print list of reductions using C's pretty printer

)
where

import FFI
import GreenCard
import Maybe
import Char(digitToInt,isDigit)

type HatTrace        = ForeignObj  -- exported

type RefNumber       = Int         -- internal only

type HatNodeInternal = (HatTrace,RefNumber)           -- internal only

data HatNode         = HatNode (HatTrace,RefNumber) | -- exported as abstract type only
		       HatInvalid

-- exported data types (exported with constructors)
data HatNodeType     = HatApplNode      | HatConstantNode |
                       HatSAT_ANode     | HatSAT_BNode    | HatSAT_CNode |
                       HatHiddenNode    | HatDummyNode    |
		       HatProjNode      | HatConstrNode   | HatIdentNode |
		       HatCaseNode      | HatLambdaNode   | HatIntNode   | HatCharNode  |
		       HatIntegerNode   | HatRationalNode | HatFloatNode | HatDoubleNode|
		       HatCStringNode   | HatIfNode       | HatGuardNode |
		       HatContainerNode | HatInvalidNode  | HatModuleNode
		       deriving (Show,Eq)

data HatApplType     = HatApplied       | HatBlackholed   | HatComplete  | HatValue
		       deriving (Show,Eq)

data HatInfixType    = HatInfix Int     | HatInfixR Int   | HatInfixL Int| HatNoInfix 
                       deriving (Show,Eq)
                       -- HatNoInfix represents default fixity

data ObserveResult   = IdentNotFound    | TopIdentNotFound| Found [HatNode]

data HatSourceRef    = HatSourceRef {row           :: Int,
				     column        :: Int,
				     moduleName    :: String,
				     moduleFile    :: String,
				     moduletrusted :: Bool} |
                       HatNoSourceRef

-- class for all structures representing Hat redexes
class Show a => HatRep a where
  toHatNode        :: a -> HatNode
  toUndefined      :: a -> a
  hatParent        :: a -> a
  hatApplFun       :: a -> a
  hatApplArgs      :: a -> [a]
  hatInfix         :: a -> HatInfixType
  hatResult        :: a -> a
  hatName          :: a -> String
  hatProjValue     :: a -> a
  hatValueInt      :: a -> Int
  hatValueChar     :: a -> Char
  hatValueInteger  :: a -> Integer
  hatValueRational :: a -> Rational
  hatValueFloat    :: a -> Float
  hatValueDouble   :: a -> Double
  hatValueString   :: a -> String
  hatNodeType      :: a -> HatNodeType
  hatSourceRef     :: a -> HatSourceRef
  hatSourceRef node = _hatSourceRef (toHatNode node)





------------------------------------------------------------------------
--
-- GREENCARD
--
------------------------------------------------------------------------
-- internal functions for conversions
------------------------------------------------------------------------



foreign import "hs_intForeignObjAddr" intForeignObjAddr :: ForeignObj -> Int

foreign import "hs_lookupAddr" lookupAddr :: Addr -> Int

foreign import "hs_lookupArray" lookupArray :: ForeignObj -> Int -> Int

_toHatNode :: HatNodeInternal -> HatNode
_toHatNode expr@(_,nodenumber) =
    if (nodenumber == 0) then HatInvalid else HatNode expr

_fromHatNode :: HatNode -> HatNodeInternal
_fromHatNode (HatNode v) = v
_fromHatNode HatInvalid = error "_fromHatNode on Invalid node!" -- should never occur



------------------------------------------------------------------------
------------------------------------------------------------------------
-- internal functions: immediately interfacing C

foreign import hs_c_openTrace :: PackedString -> IO ForeignObj

c_openTrace :: String -> IO HatTrace
c_openTrace tmp1 =
  let name = toCString tmp1
  in do hs_c_openTrace name



foreign import "hs_c_isInvalidTrace" c_isInvalidTrace :: ForeignObj -> Int

foreign import "hs_c_hatMain" c_hatMain :: ForeignObj -> Int

foreign import "hs_c_hatFirst" c_hatFirst :: ForeignObj -> Int

foreign import hs_c_hatNext :: ForeignObj -> Int -> Int

c_hatNext :: HatNodeInternal -> RefNumber
c_hatNext (hattrace,nodenumber) =
  let newnode = hs_c_hatNext hattrace nodenumber
  in newnode



foreign import hs_c_hatParent :: ForeignObj -> Int -> Int

c_hatParent :: HatNodeInternal -> RefNumber
c_hatParent (hattrace,nodenumber) =
  let newnode = hs_c_hatParent hattrace nodenumber
  in newnode



------------------------------------------------
-- interface to source reference and module names

-- get node within hatfile, containing SrcRef
foreign import hs_c_hatSrcRefNode :: ForeignObj -> Int -> Int

c_hatSrcRefNode :: HatNodeInternal -> RefNumber
c_hatSrcRefNode (hattrace,nodenumber) =
  let newnode = hs_c_hatSrcRefNode hattrace nodenumber
  in newnode



-- get module info node
foreign import hs_c_hatModInfo :: ForeignObj -> Int -> Int

c_hatModInfo :: HatNodeInternal -> RefNumber
c_hatModInfo (hattrace,nodenumber) =
  let newnode = hs_c_hatModInfo hattrace nodenumber
  in newnode



foreign import hs_c_hatPosnRow :: ForeignObj -> Int -> Int

c_hatPosnRow :: HatNodeInternal -> Int
c_hatPosnRow (hattrace,nodenumber) =
  let value = hs_c_hatPosnRow hattrace nodenumber
  in value



foreign import hs_c_hatPosnColumn :: ForeignObj -> Int -> Int

c_hatPosnColumn :: HatNodeInternal -> Int
c_hatPosnColumn (hattrace,nodenumber) =
  let value = hs_c_hatPosnColumn hattrace nodenumber
  in value



foreign import hs_c_hatModName :: ForeignObj -> Int -> PackedString

c_hatModName :: HatNodeInternal -> String
c_hatModName (hattrace,nodenumber) =
  let value = hs_c_hatModName hattrace nodenumber
  in (fromCString value)



foreign import hs_c_hatModSrcName :: ForeignObj -> Int -> PackedString

c_hatModSrcName :: HatNodeInternal -> String
c_hatModSrcName (hattrace,nodenumber) =
  let value = hs_c_hatModSrcName hattrace nodenumber
  in (fromCString value)



foreign import hs_c_hatModSrcTrusted :: ForeignObj -> Int -> Int

c_hatModSrcTrusted :: HatNodeInternal -> Int
c_hatModSrcTrusted (hattrace,nodenumber) =
  let value = hs_c_hatModSrcTrusted hattrace nodenumber
  in value




----------------------------------------------------

foreign import hs_c_hatResult :: ForeignObj -> Int -> Int

c_hatResult :: HatNodeInternal -> RefNumber
c_hatResult (hattrace,nodenumber) =
  let newnode = hs_c_hatResult hattrace nodenumber
  in newnode



foreign import hs_c_hatLeftmost :: ForeignObj -> Int -> Int

c_hatLeftmost :: HatNodeInternal -> RefNumber
c_hatLeftmost (hattrace,nodenumber) =
  let newnode = hs_c_hatLeftmost hattrace nodenumber
  in newnode



foreign import hs_c_hatName :: ForeignObj -> Int -> PackedString

c_hatName :: HatNodeInternal -> String
c_hatName (hattrace,nodenumber) =
  let s = hs_c_hatName hattrace nodenumber
  in (fromCString s)



foreign import hs_c_hatInt :: ForeignObj -> Int -> Int

c_hatInt :: HatNodeInternal -> Int
c_hatInt (hattrace,nodenumber) =
  let i = hs_c_hatInt hattrace nodenumber
  in i



foreign import hs_c_hatInteger :: ForeignObj -> Int -> Int

c_hatInteger :: HatNodeInternal -> Int
c_hatInteger (hattrace,nodenumber) =
  let i = hs_c_hatInteger hattrace nodenumber
  in i




foreign import hs_c_hatChar :: ForeignObj -> Int -> Char

c_hatChar :: HatNodeInternal -> Char
c_hatChar (hattrace,nodenumber) =
  let c = hs_c_hatChar hattrace nodenumber
  in c




foreign import hs_c_hatDouble :: ForeignObj -> Int -> Double

c_hatDouble :: HatNodeInternal -> Double
c_hatDouble (hattrace,nodenumber) =
  let d = hs_c_hatDouble hattrace nodenumber
  in d




foreign import hs_c_hatRational :: ForeignObj -> Int -> Double

c_hatRational :: HatNodeInternal -> Double
c_hatRational (hattrace,nodenumber) =
  let r = hs_c_hatRational hattrace nodenumber
  in r




foreign import hs_c_hatFloat :: ForeignObj -> Int -> Float

c_hatFloat :: HatNodeInternal -> Float
c_hatFloat (hattrace,nodenumber) =
  let fl = hs_c_hatFloat hattrace nodenumber
  in fl



foreign import hs_c_hatCString :: ForeignObj -> Int -> PackedString

c_hatCString :: HatNodeInternal -> String
c_hatCString (hattrace,nodenumber) =
  let s = hs_c_hatCString hattrace nodenumber
  in (fromCString s)




foreign import hs_c_hatApplFun :: ForeignObj -> Int -> Int

c_hatApplFun :: HatNodeInternal -> RefNumber
c_hatApplFun (hattrace,nodenumber) =
  let newnode = hs_c_hatApplFun hattrace nodenumber
  in newnode



foreign import hs_c_hatProjRef :: ForeignObj -> Int -> Int

c_hatProjRef :: HatNodeInternal -> RefNumber
c_hatProjRef (hattrace,nodenumber) =
  let newnode = hs_c_hatProjRef hattrace nodenumber
  in newnode



foreign import hs_c_hatApplArity :: ForeignObj -> Int -> Int

c_hatApplArity :: HatNodeInternal -> Int
c_hatApplArity (hattrace,nodenumber) =
  let newnode = hs_c_hatApplArity hattrace nodenumber
  in newnode



foreign import hs_gethatArgInternal :: ForeignObj -> Int -> Int -> Int

gethatArgInternal :: HatNodeInternal -> Int -> Int
gethatArgInternal (hattrace,nodenumber) argnum =
  let newnode = hs_gethatArgInternal hattrace nodenumber argnum
  in newnode



-- functions accessing higher level C functions
foreign import hs_c_hatIsCAF :: ForeignObj -> Int -> Int

c_hatIsCAF :: HatNodeInternal -> Int
c_hatIsCAF (hattrace,nodenumber) =
  let value = hs_c_hatIsCAF hattrace nodenumber
  in value



foreign import hs_c_hatIsTopLevel :: ForeignObj -> Int -> Int

c_hatIsTopLevel :: HatNodeInternal -> Int
c_hatIsTopLevel (hattrace,nodenumber) =
  let value = hs_c_hatIsTopLevel hattrace nodenumber
  in value



foreign import hs_c_hatIsTrusted :: ForeignObj -> Int -> Int

c_hatIsTrusted :: HatNodeInternal -> Int
c_hatIsTrusted (hattrace,nodenumber) =
  let value = hs_c_hatIsTrusted hattrace nodenumber
  in value



foreign import hs_c_hatExpressionStr :: ForeignObj -> Int -> Int -> Int -> PackedString

c_hatExpressionStr :: HatNodeInternal -> Int -> Int -> String
c_hatExpressionStr (hattrace,nodenumber) verbose precision =
  let prettystring = hs_c_hatExpressionStr hattrace nodenumber verbose precision
  in (fromCString prettystring)




-- converting C ID value to Haskell type
toHatNodeType :: Int -> HatNodeType
toHatNodeType  0 = HatApplNode
toHatNodeType  1 = HatConstantNode
toHatNodeType  2 = HatProjNode
toHatNodeType  3 = HatHiddenNode
toHatNodeType  4 = HatSAT_ANode
toHatNodeType  5 = HatSAT_BNode
toHatNodeType  6 = HatSAT_CNode
toHatNodeType 12 = HatSAT_ANode
toHatNodeType 13 = HatSAT_BNode
toHatNodeType 14 = HatSAT_CNode
toHatNodeType 32 = HatModuleNode
toHatNodeType 64 = HatIntNode
toHatNodeType 65 = HatCharNode
toHatNodeType 66 = HatIntegerNode
toHatNodeType 67 = HatRationalNode
toHatNodeType 68 = HatFloatNode
toHatNodeType 69 = HatDoubleNode
toHatNodeType 70 = HatIdentNode
toHatNodeType 71 = HatConstrNode
toHatNodeType 74 = HatCaseNode
toHatNodeType 75 = HatLambdaNode
toHatNodeType 76 = HatDummyNode
toHatNodeType 77 = HatCStringNode
toHatNodeType 78 = HatIfNode
toHatNodeType 79 = HatGuardNode
toHatNodeType 80 = HatContainerNode
toHatNodeType 86 = HatIdentNode
toHatNodeType _  = HatInvalidNode

fromHatNodeType _ = 0

foreign import hs_c_hatNodeType :: ForeignObj -> Int -> Int

c_hatNodeType :: HatNodeInternal -> HatNodeType
c_hatNodeType (hattrace,nodenumber) =
  let nodetype = hs_c_hatNodeType hattrace nodenumber
  in (toHatNodeType nodetype)



-- converting C ID values to Haskell type
toHatApplType 4 = Just HatApplied
toHatApplType 5 = Just HatBlackholed
toHatApplType 6 = Just HatComplete
toHatApplType 71 = Just HatValue
toHatApplType _ = Nothing

fromHatApplType _ = 0

foreign import hs_c_hatApplType :: ForeignObj -> Int -> Int

c_hatApplType :: HatNodeInternal -> Maybe HatApplType
c_hatApplType (hattrace,nodenumber) =
  let nodetype = hs_c_hatApplType hattrace nodenumber
  in (toHatApplType nodetype)



foreign import hs_c_hatInfixValue :: ForeignObj -> Int -> Int

c_hatInfixValue :: HatNodeInternal -> Int
c_hatInfixValue (hattrace,nodenumber) =
  let infixprio = hs_c_hatInfixValue hattrace nodenumber
  in infixprio




--------------------------------------------------------------------------
-- interfacing detect.c

foreignObjToList :: HatTrace -> ForeignObj -> [HatNode]
foreignObjToList hattrace dat =
    foreignObjToList' 0
    where foreignObjToList' i =
            let v = (lookupArray dat i) in
    	     if (v == -1) then [] else
		    (HatNode (hattrace,v)):(foreignObjToList' (i+1))

foreign import hs_getEDTForeignObj :: ForeignObj -> Int -> ForeignObj

getEDTForeignObj :: HatNodeInternal -> ForeignObj
getEDTForeignObj (hattrace,nodenumber) =
  let resultArray = hs_getEDTForeignObj hattrace nodenumber
  in resultArray



--------------------------------------------------------------------------
-- Interfacing observe.c

-- submit query to c (observing a function)
foreign import hs_getObserveInt :: ForeignObj -> PackedString -> PackedString -> Int -> ForeignObj

getObserveInt :: HatTrace -> String -> String -> Int -> ForeignObj
getObserveInt hattrace tmp1 tmp2 recursive =
  let ident = toCString tmp1
      topIdent = toCString tmp2
      query = hs_getObserveInt hattrace ident topIdent recursive
  in query



-- submit query to c (observing a source reference)
foreign import hs_getObserveSrcInt :: ForeignObj -> PackedString -> Int -> Int -> ForeignObj

getObserveSrcInt :: HatTrace -> String -> Int -> Int -> ForeignObj
getObserveSrcInt hattrace tmp1 line column =
  let moduleName = toCString tmp1
      query = hs_getObserveSrcInt hattrace moduleName line column
  in query



-- submit query to c (observing all observable)
foreign import "hs_getObservablesInt" getObservablesInt :: ForeignObj -> ForeignObj

-- get one element from the c result
foreign import "hs_getObservedElement" getObservedElement :: ForeignObj -> Int -> Int

-- get status of query: identifier found?
foreign import "hs_getObserveIntStatus" getObserveIntStatus :: ForeignObj -> Int

-- convert list of c results in haskell list
foreignObserveToList :: HatTrace -> ForeignObj -> [HatNode]
foreignObserveToList hattrace f =
  foreignObserveToList' 0
  where
   foreignObserveToList' i =
     let x = (getObservedElement f i) in
       if (x==0) then [] else (HatNode (hattrace,x)):(foreignObserveToList' (i+1))



--------------------------------------------------------------------------
--                                                                      --
--                    Exported Functions                                --
--                                                                      --
--------------------------------------------------------------------------



--------------------------------------------------------------------------
-- define Eq and Show instances for HatNode

instance Eq HatNode where
  HatInvalid == HatInvalid = True
  (HatNode (a,b)) == (HatNode (c,d)) =
      ((intForeignObjAddr a)==(intForeignObjAddr c))&&(b==d)
  _ == _ = False

instance Show HatNode where
  show HatInvalid  = "Invalid"
  show (HatNode (hattrace,node)) =
      (show (intForeignObjAddr hattrace))++":"++(show node)
--------------------------------------------------------------------------

-- predicates for HatNode
isValidNode   :: HatRep a => a -> Bool
isValidNode node = (hatNodeType node)/=HatInvalidNode

isInvalidNode :: HatRep a => a -> Bool
isInvalidNode node = (hatNodeType node)==HatInvalidNode

hatInvalidNode = HatInvalid  -- don't export structure of HatNode, just a value!

-- convert any HatRep to external representation (in order to pass it as an command-line
-- argument to another hat tool).
toRemoteRep :: HatRep a => a -> String
toRemoteRep exp = if (isValidNode exp) then
		    let (_,node)=(_fromHatNode (toHatNode exp)) in 
				  show node
		   else
                    "INVALID"

-- convert  external representation to HatNode (in order to accept an command-line
-- argument from another hat tool).
fromRemoteRep :: HatTrace -> String -> HatNode
fromRemoteRep hattrace s = let i = stoi s in
				  if (isNothing i) then HatInvalid else
					 (HatNode (hattrace,(fromJust i)))
  where stoi s = stoi' 0 s
	stoi' i [] =  Just i
	stoi' i (c:r) | (isDigit c) = stoi' (i*10+(digitToInt c)) r
		      | otherwise = Nothing

-- open a trace file
openTrace :: String -> IO (Maybe HatTrace)
openTrace s =
    do
      h <- c_openTrace s
      if ((c_isInvalidTrace h)/=0) then return Nothing else return (Just h)

-- get name of a trace file
foreign import hs_hatTraceName :: ForeignObj -> PackedString

hatTraceName :: HatTrace -> String
hatTraceName hattrace =
  let s = hs_hatTraceName hattrace
  in (fromCString s)



-- get the node type of a HatNode (exportet in HatRep instance for HatNode)
_hatNodeType :: HatNode -> HatNodeType
_hatNodeType (HatNode expr) = c_hatNodeType expr
_hatNodeType _              = HatInvalidNode

-- get the parent for a HatNode (exportet in HatRep instance for HatNode)
_hatParent :: HatNode -> HatNode
_hatParent (HatNode expr@(hattrace,nodenumber)) = _toHatNode (hattrace,c_hatParent expr)
_hatParent _ = error "Call to \"hatParent\" with \"Invalid\" node."

-- get the original parent of a projection node (former indirection)
_hatProjRef :: HatNode -> HatNode
_hatProjRef (HatNode expr@(hattrace,nodenumber)) = _toHatNode (hattrace,c_hatProjRef expr)
_hatProjRef _ = error "Call to \"hatProjRef\" with \"Invalid\" node."

-- get the name of an HatName node (exportet in HatRep instance for HatNode)
_hatName :: HatNode -> Maybe String
_hatName (HatNode hatexpr) = 
 let s = (c_hatName hatexpr) in
  if (s=="") then Nothing else (Just s)
_hatName _ = error "Call to \"hatName\" with \"Invalid\" node."

-- get the node representing the function of an application node
_hatApplFun :: HatNode -> HatNode
_hatApplFun (HatNode expr@(hattrace,nodenumber)) = _toHatNode (hattrace,c_hatApplFun expr)
_hatApplFun _ = error "Call to \"hatApplFun\" with \"Invalid\" node."

-- return arity of an application node (exportet in HatRep instance for HatNode)
_hatApplArity :: HatNode -> Int
_hatApplArity = c_hatApplArity . _fromHatNode

-- get the list of arguments of an application node
_hatApplArgs :: HatNode -> [HatNode]
_hatApplArgs (HatNode expr@(hattrace,h)) = gethatArgs 0 (c_hatApplArity expr)
 where gethatArgs _ 0 = []
       gethatArgs i j = (HatNode (hattrace,(gethatArgInternal expr i))):
			(gethatArgs (i+1) (j-1))
_hatApplArgs _ = error "Call to \"hatApplArgs\" with \"Invalid\" node."

-- get infix priority and type of an application node
_hatInfix :: HatNode -> HatInfixType
_hatInfix (HatNode expr) = 
  let infixP = (c_hatInfixValue expr) in 
   toInfixType' (infixP `mod` 4) (infixP `div` 4)
  where
   toInfixType' 0 infixP = HatInfix infixP
   toInfixType' 1 infixP = HatInfixR infixP
   toInfixType' 2 infixP = HatInfixL infixP
   toInfixType' _ _      = HatNoInfix
_hatInfix _ = error "Call to \"hatInfix\" with \"Invalid\" node."

-- get type of application (Applied,Blackholed,Complete,Value) 
_hatApplType :: HatNode -> Maybe HatApplType
_hatApplType = c_hatApplType . _fromHatNode

-- get result of an application (if application of type complete, otherwise Nothing)
_hatResult :: HatNode -> HatNode
_hatResult (HatNode expr@(hattrace,nodenumber)) = _toHatNode (hattrace,c_hatResult expr)
_hatResult _ = error "Call to \"hatResult\" with \"Invalid\" node."

-- get information about source position, for Identifier/Constructor/Application/Name
-- nodes
-- returns a HatSourceRef (RowPosition,ColumnPosition,ModuleName,ModuleFile,Trusted)
_hatSourceRef :: HatNode -> HatSourceRef
_hatSourceRef (HatNode expr@(hattrace,nodenumber)) =
    let srcrefNode = (hattrace,c_hatSrcRefNode expr);
	modinfoNode = if ((snd srcrefNode)==0) then (hattrace,0) else 
                        (hattrace,c_hatModInfo srcrefNode) in
      if ((snd modinfoNode)==0) then HatNoSourceRef else
	(HatSourceRef
	 (c_hatPosnRow srcrefNode)
	 (c_hatPosnColumn srcrefNode)
	 (c_hatModName modinfoNode)
	 (c_hatModSrcName modinfoNode)
	 ((c_hatModSrcTrusted modinfoNode)==1))
_hatSourceRef _ = error "Call to \"hatSourceRef\" with \"Invalid\" node."

-- get node representing the "main" caf
hatMain :: HatTrace -> HatNode
hatMain hattrace = _toHatNode (hattrace,c_hatMain hattrace)

-- get the first node within the HatTrace file
hatTraceFirst :: HatTrace -> HatNode
hatTraceFirst hattrace = _toHatNode (hattrace,c_hatFirst hattrace)

-- get node immediately following a node
hatTraceNext :: HatRep a => a -> HatNode
hatTraceNext exp =
    let node = toHatNode exp;
	HatNode (hattrace,_) = node in
	if (isInvalidNode node) then HatInvalid else
	   _toHatNode (hattrace,c_hatNext (_fromHatNode node))

-- get leftmost outermost symbol of an application (returns an identifier/constructor)
hatLeftmost :: HatRep a => a -> HatNode
hatLeftmost exp =
    let node = (toHatNode exp) in
      if (isValidNode node) then
        let (HatNode expr@(hattrace,nodenumber)) = node in
	    _toHatNode (hattrace,c_hatLeftmost expr)
       else error "Call to \"hatLeftmost\" with \"Invalid\" node."

hatIsCAF      :: HatRep a => a -> Bool
hatIsCAF exp =
    let node = (toHatNode exp) in
      if (isValidNode node) then (c_hatIsCAF (_fromHatNode node)) == 1
	 else error "Call to \"hatIsCAF\" with \"Invalid\" node."

hatIsTopLevel :: HatRep a => a -> Bool
hatIsTopLevel exp =
    let node = (toHatNode exp) in
      if (isValidNode node) then (c_hatIsTopLevel (_fromHatNode node)) == 1
	 else error "Call to \"hatIsTopLevel\" with \"Invalid\" node."

hatIsTrusted  :: HatRep a => a -> Bool
hatIsTrusted exp =
    let node = (toHatNode exp) in
     if (isValidNode node) then (c_hatIsTrusted (_fromHatNode node))==1
	else error "Call to \"hatIsTrusted\" with \"Invalid\" node."

--------------------------------------------------------------------------
-- EDT interface

-- returns a list of children for a node
detect :: HatRep a => a -> [HatNode]
detect exp =
    let node = (toHatNode exp) in
     if (isValidNode node) then
       let (HatNode e@(hattrace,_)) = node in
	   (foreignObjToList hattrace (getEDTForeignObj e))
      else error "Call to \"detect\" with \"Invalid\" node."

--------------------------------------------------------------------------
-- observational interface

-- returns observations within hattrace: all applications of "ident", within
--  a function "topIdent" (leave empty otherwise). If recursive is true, recursive
--  calls to the application are omitted
observe :: HatTrace -> String -> String -> Bool -> ObserveResult
observe hattrace ident topIdent recursive =
  let x = (getObserveInt hattrace ident topIdent (if recursive then 1 else 0));
      s = (getObserveIntStatus x) in
   if ((s==2)&&((null topIdent)==False)) then TopIdentNotFound else
    if (s==1) then IdentNotFound else
     (Found (foreignObserveToList hattrace x))

-- returns observations within hattrace: all applications/values at the given
-- source position are returned. If moduleName is an empty string, the main module
-- is considered.
observeSrc :: HatTrace -> String -> Int -> Int -> ObserveResult
observeSrc hattrace moduleName line column =
  let x = (getObserveSrcInt hattrace moduleName line column) in
     (Found (foreignObserveToList hattrace x))

-- returns observations within hattrace: all applications/values at the given
-- source position are returned. If moduleName is an empty string, the main module
-- is considered.
observables :: HatTrace -> ObserveResult
observables hattrace =
  let x = (getObservablesInt hattrace) in
     (Found (foreignObserveToList hattrace x))

-- predicates for result returned by observe

isIdentNotFound :: ObserveResult       -> Bool
isIdentNotFound    IdentNotFound        = True
isIdentNotFound _                       = False

isTopIdentNotFound :: ObserveResult    -> Bool
isTopIdentNotFound    TopIdentNotFound  = True
isTopIdentNotFound _                    = False

isFound :: ObserveResult   -> Bool
isFound (Found _)           = True
isFound _                   = False

fromFound :: ObserveResult -> [HatNode]
fromFound (Found r)         = r
fromFound _                 = []

------------------------------------------------------------------------------
-- pretty printing by C (haskell version in module "HatExpression" available

hatCExpressionStr :: HatRep a => Bool -> Int -> a -> String
hatCExpressionStr verbose precision node =
    c_hatExpressionStr (_fromHatNode (toHatNode node))
			   (if verbose then 1 else 0)
			   precision

printCExpression :: HatRep a => Int -> a -> IO ()
printCExpression precision exp =
    let node = (toHatNode exp) in
    if (isInvalidNode node) then putStr "Nothing" else
      putStr (hatCExpressionStr False precision node)

printCReduction :: HatRep a => Int -> a -> IO ()
printCReduction precision exp =
  printCExpression precision exp >>
  if (isValidNode exp) then
    let res = (hatResult exp) in
    if (isValidNode res) then
       putStr " = " >>
       (printCExpression precision res)
     else
      return ()
  else
   return ()

printCReductionList :: HatRep a => Int-> [a] -> IO ()
printCReductionList _ [] = return ()
printCReductionList precision (e:list) =
   do
    printCReduction precision e
    putStrLn ""
    printCReductionList precision list

------------------------------------------------------------------------------


instance HatRep HatNode where
  toHatNode node   = node
  toUndefined _    = HatInvalid
  hatParent        = _hatParent
  hatApplFun       = _hatApplFun
  hatApplArgs      = _hatApplArgs
  hatResult        = _hatResult
  hatName          = fromJust . _hatName
  hatInfix         = _hatInfix
  hatProjValue     = _hatProjRef
  hatValueInt      = c_hatInt . _fromHatNode
  hatValueChar     = c_hatChar . _fromHatNode
  hatValueInteger  = toInteger . c_hatInteger . _fromHatNode
  hatValueRational = toRational . c_hatRational . _fromHatNode
  hatValueFloat    = c_hatFloat . _fromHatNode
  hatValueDouble   = c_hatDouble . _fromHatNode
  hatValueString   = c_hatCString . _fromHatNode
  hatNodeType      = _hatNodeType




